<?php

require_once DRUPAL_ROOT . '/includes/cache.inc';

class SchrodiCacheDatabaseCache extends DrupalDatabaseCache {
  /**
   * Implements DrupalCacheInterface::set().
   */
  function set($cid, $data, $expire = CACHE_PERMANENT) {
    $fields = array(
      'serialized' => 0,
      'created' => microtime(TRUE),
      'expire' => $expire,
    );
    if (!is_string($data)) {
      $fields['data'] = serialize($data);
      $fields['serialized'] = 1;
    }
    else {
      $fields['data'] = $data;
      $fields['serialized'] = 0;
    }
    try {
      db_merge($this->bin)
        ->key(array('cid' => $cid))
        ->fields($fields)
        ->execute();
    }
    catch (Exception $e) {
      // The database may not be available, so we'll ignore cache_set requests.
    }
  }
}

class SchrodiCacheCache implements DrupalCacheInterface {

  /**
   * Prefix for bin-specific entry to track the last write.
   */
  const LAST_WRITE_TIMESTAMP_PREFIX = 'cache_last_write_timestamp_';

  protected $lastWriteTimestamp;

  /**
   * The consistent cache backend.
   */
  protected $consistentBackend;

  /**
   * Constructs a DrupalDatabaseCache object.
   *
   * @param $bin
   *   The cache bin for which the object is created.
   */
  function __construct($bin) {
    global $conf;

    $class = 'SchrodiCacheDatabaseCache';
    if (!empty($conf['schrodicache']['consistent_backend'])) {
      $class = $conf['schrodicache']['consistent_backend'];
    }

    $this->consistentBackend = new $class($bin);
    $this->bin = $bin;
    $this->apcPrefix = $bin;
  }

  /**
   * Implements DrupalCacheInterface::isEmpty().
   */
  function isEmpty() {
    return $this->consistentBackend->isEmpty();
  }

  /**
   * Implements DrupalCacheInterface::get().
   */
  function get($cid) {
    if ($last_write_timestamp = $this->lastWriteTimestamp()) {
      $item = apc_fetch($this->apcPrefix . $cid);
      if ($item && $item->created >= $last_write_timestamp) {
        return $item;
      }
    }

    if ($item = $this->consistentBackend->get($cid)) {
      $this->apcSet($item->cid, $item->data);
    }
    return $item;
  }

  /**
   * Implements DrupalCacheInterface::get().
   */
  function set($cid, $data, $expire = CACHE_PERMANENT) {
    $this->markAsOutdated();
    $this->consistentBackend->set($cid, $data, $expire);
    $this->apcSet($cid, $data);
  }

  function apcSet($cid, $data, $expire = CACHE_PERMANENT) {
    $fields = (object) array(
      'serialized' => 0,
      'created' => REQUEST_TIME,
      'expire' => $expire,
      'cid' => $cid,
      'data' => $data,
      'serialized' => 0,
    );
    apc_store($this->apcPrefix . $cid, $fields);
  }

  /**
   * Implements DrupalCacheInterface::clear().
   */
  function clear($cid = NULL, $wildcard = FALSE) {
    $this->consistentBackend->clear($cid, $wildcard);
    if (!empty($cid)) {
      if ($wildcard) {
        if ($cid == '*') {
          apc_delete(new APCIterator('user', '/^' . $this->apcPrefix . '/'));
        }
        else {
          apc_delete(new APCIterator('user', '/^' . $this->apcPrefix . $cid . '/'));
        }
      }
      else {
        apc_delete($cid);
      }
    }
    else {
      apc_delete(new APCIterator('user', '/^' . $this->apcPrefix . '/'));
    }
  }

  /**
   * Implements DrupalCacheInterface::getMultiple().
   */
  function getMultiple(&$cids) {
    $cache = array();
    if ($last_write_timestamp = $this->lastWriteTimestamp()) {
      $prefixed_cids = array();
      foreach ($cids as $cid) {
        $prefixed_cids[$this->apcPrefix . $cid] = $cid;
      }
      $result = apc_fetch(array_keys($prefixed_cids));
      foreach ($result as $item) {
        if ($item->created >= $last_write_timestamp) {
          $cache[$item->cid] = $item;
        }
      }
      $cids = array_diff($cids, array_keys($cache));
    }

    if ($cids) {
      $items = $this->consistentBackend->getMultiple($cids);
      foreach ($items as $item) {
        $cache[$item->cid] = $item;
        $this->apcSet($item->cid, $item->data);
      }
    }

    return $cache;
  }

  /**
   * Initialize the last write timestamp using the state instance.
   */
  public function lastWriteTimestamp() {
    if ($this->lastWriteTimestamp === NULL) {
      $result = db_query('SELECT value FROM {variable} WHERE name = :name', array(':name' => 'schrodicache_last_write_timestamp'))->fetchField();
      if ($result) {
        $this->lastWriteTimestamp = unserialize($result);
      }
      else {
        $this->lastWriteTimestamp = microtime(TRUE);
        $this->markAsOutdated();
      }
    }
    return $this->lastWriteTimestamp;
  }

  protected function setLastWriteTimestamp($timestamp) {
    $this->lastWriteTimestamp = $timestamp;
  }

  /**
   * Mark the inconsistent cache bin as outdated because of a write.
   */
  protected function markAsOutdated() {
    $now = microtime(TRUE);
    $this->setLastWriteTimestamp($now);
    variable_set('schrodicache_last_write_timestamp', $now);
  }
}

